#!/bin/sh
# \
exec /usr/bin/adedit "$0" "$@"


proc const {constvar} { return $constvar }
proc def {name = value} {interp alias {} $name {} const $value}

set BIN_INSTALL "/usr"
set CIMS_INSTALL "/usr/share/centrifydc"

def EXIT_SUCCESS        = 0
def EXIT_ERR_ROOT_PRIV  = 1 
def EXIT_ERR_USAGES     = 2 
def EXIT_ERR_INVALID_OUTPUT_FILE  = 3
def EXIT_ERR_NO_CIP_INFO = 4
def EXIT_ERR_KRB5CCACHE = 5
def EXIT_ERR_GETTMPDIR = 6
def EXIT_ERR_NOT_JOINED = 7
def EXIT_ERR_WRONG_USER_NAME_FORMAT = 8
 
 
# adinfo return code.
def ADINFO_ERR_NOT_JOINED = 10
def ADINFO_ERR_CIP_DIRECT_CONNECT = 180
def ADINFO_ERR_CIP_HTTP_PROXY_CONF = 181
def ADINFO_ERR_CIP_CC_PROXY_ADCLIENT = 182
def ADINFO_ERR_CIP_CC_PROXY_KSET = 183
def ADINFO_ERR_CIP_NO_INFO = 184

def ADWEBPROXYCONF_ERROR_NO_FILE_FOUND = 158

def MACHINE_ACCOUNT  = "S-1-5-18"

def NAME_AUTH_NONE      = "none"
def NAME_AUTH_NEGOTIATE = "negotiate"
def NAME_AUTH_ANY       = "anyauth"
def NAME_AUTH_BASIC     = "basic"
def NAME_AUTH_DIGEST    = "digest"
def NAME_AUTH_NTLM      = "ntlm"

def ADWEBPROXYCONF_CMD  = "$BIN_INSTALL/bin/adwebproxyconf"
def ADINFO_CMD = "$BIN_INSTALL/bin/adinfo"
def KINIT_CMD = "$CIMS_INSTALL/kerberos/bin/kinit"
def CURL_CMD = "$CIMS_INSTALL/bin/curl"
def ADID_CMD = "$CIMS_INSTALL/bin/adid"


# Global variable
set cip ""
set TestConnectionTimeout 5
set DownloadRESTAPI "servermanage/getmastersshkey?format=pem"

array set WebProxy {
    DirectConnect       false
    Server              "" 
    AuthRequred         false
    AuthType            ""
    AuthUser            ""
    AuthKrb5CCache      ""
}

array set AuthTypeProxySupported {
    NAME_AUTH_NEGOTIATE 0
    NAME_AUTH_DIGEST    0
    NAME_AUTH_NTLM      0
    NAME_AUTH_BASIC     0
}

proc GetTempDir {} {
    if {[catch {set temp_dir [string trim [get_temp_dir]]} err_msg]} {
        puts "$err_msg"
        set temp_dir ""
    }
    return $temp_dir    
}

set SecureTempDir    [GetTempDir]
if {$SecureTempDir eq ""} {
    exit [expr [EXIT_ERR_GETTMPDIR]]
}

proc usage {} {
    append tool_usage "This tool is used to download ssh public key of associated tenant on Unix machines which has DA installed.  Usage:\n"
    append tool_usage "Synopsis:\n"
    append tool_usage "dadownloadsshpublickey.tcl \[--cip cip] \[--output-file file]\n"
    append tool_usage "Options:\n"
    append tool_usage "    -i, --cip <cip url>\n"
    append tool_usage "        The cip url which is used to compose the ssh public key REST API URL.\n"
    append tool_usage "        If not specified, will automatically get it.\n"
    append tool_usage "    -o, --output-file <file>\n"
    append tool_usage "        Write the ssh public key to file.\n"
    append tool_usage "Exit status:\n"
    append tool_usage "  0           success\n"
    append tool_usage "  non-zero    otherwise\n"
    puts $tool_usage
}


#Command line Option setting
set OptionConfig {
    "-i" {
        HASARG true
        VNAME  "cip"
    }
    "-o" {
        HASARG true
        VNAME  "OutPutFile"
    }
}

set OptionMap {
    "--cip"            "-i"
    "--output-file"    "-o" 
}


# Execute Shell Command
#
# Inputs:    command to be executed
#
# return:    $status    command result code
#            $output    command output

proc ExecuteCommand {command} {
    set status 0
    set outputs ""
    if {[catch {eval exec $command} outputs states]} {
    # states : -code 1 -level 0 -errorcode {CHILDSTATUS pid rc} -errorinfo {error messages} -errorline 1
        set details [dict get $states -errorcode]
        if {[lindex $details 0] eq "CHILDSTATUS"} {
            set status [lindex $details 2]
        } else {
            puts "Unexpected Status when executing command '$command': $states"
            set status [dict get $states -code]
        }
    }
    return [list $status $outputs]    
}



# Get the path of krb5 ccache
# If it fails in reading "adclient.krb5.ccache.file", it uses the default value.
proc GetKrb5CCachePath {} {
    set krb5CcachePath ""

    lassign [ExecuteCommand "[ADINFO_CMD] -c | grep adclient.krb5.ccache.file"] rc output
    if {$rc == 0 && [regexp {(.*)?: *(.*)} $output -> -> value]} {
       set krb5CcachePath $value
    }
    
    if {$krb5CcachePath eq ""} {
        return "/etc/krb5.ccache"
    } else {
        return $krb5CcachePath
    }
}

set UserKrb5CCacheName $SecureTempDir/dadownloadsshpubkey_krb5cc_[pid]
set MachineKrb5CCacheName [GetKrb5CCachePath]


proc GetOptions {options} {
    global OptionConfig
    global OptionMap

    set count 0
    set arglen [llength $options]
    while {$count < $arglen} {

        set opt [lindex $options $count]
        if {[dict exist $OptionMap $opt]} {
            set opt [dict get $OptionMap $opt]
        }
        if {![dict exist $OptionConfig $opt]} {
            return false
        }

        if {[dict get $OptionConfig $opt HASARG]} {
            if {[expr $count + 1] >= $arglen } {
                return false
            } else {
                incr count
                set value [lindex $options $count]
            }
        } else {
            set value [dict get $OptionConfig $opt VALUE]
        }

        set cmd  "set [dict get $OptionConfig $opt VNAME] \"$value\""
        uplevel "eval $cmd"
        incr count
    }
    return true
}


proc IsOutPutFileValid {} {
    global OutPutFile
    
    set message ""
    set parentDir ""
    set fileName ""
    set rc -1
    
    if {[regexp {(.*)\/(\w+)} $OutPutFile -> parentDir fileName]} {
    
        # Check whether parent dir is a directory
        if {![file isdirectory $parentDir]} {
            set message "$parentDir isn't a directory"
            return [list $rc $message]
        }
        
        # Check if parent dir's owner is root
        if {![file owned $parentDir]} {
            set message "$parentDir is not owned by root"
            return [list $rc $message]
        }
        
        # Check if parent dir is writen by root only
        set permission [file attributes $parentDir -permissions]
        if {[expr $permission & 02] || [expr $permission & 020]} {
            set message "$parentDir is not writable by root only"
            return [list $rc $message]
        }
        
        # Check if parent dir is a symlink
        if {![catch {file readlink $parentDir}]} {
            set message "$parentDir is a symlink"
            return [list $rc $message]
        } 
        
        # Check if the file exists
        if {[file exists $OutPutFile]} {
            # it exists, then check if it is a regular file
            set type [file type $OutPutFile]
            if {$type ne "file"} {
                set message "$OutPutFile is not a regular file"
                return [list $rc $message]
            }
        }
        
        set rc 0
        return [list $rc $message]
    } else {
        set message "Output file path is invalid, "
        return [list $rc $message]
    }
    
}


# Check whether username is a machine account
proc IsMachineAccount {username} {
    if {$username eq [MACHINE_ACCOUNT]} {
        return true
    }
    return false
}


# Get proxy Krb5 credential cache setting string
# Input:     userCred    
# Output:    string of proxy Krb5 credential cache setting
proc GetProxyKrb5CCache {userCred} {
    global MachineKrb5CCacheName
    global UserKrb5CCacheName

    if {[IsMachineAccount $userCred]} {
        return $MachineKrb5CCacheName
    }

    if {[file exists $UserKrb5CCacheName]} {
        file delete $UserKrb5CCacheName
    }

    set cmd "[ADWEBPROXYCONF_CMD] -C -p |& [KINIT_CMD] -c FILE:$UserKrb5CCacheName $userCred"

    lassign [ExecuteCommand "$cmd 2>@1"] rc output
    if {$rc != 0} {
        return ""
    } 
    return $UserKrb5CCacheName
}



# Destroy user KRB5 credential cache
# Output:    $result        - 0: successful    1: failed
proc DestroyUserKrb5CCache {} {
    global UserKrb5CCacheName

    if {[file exist $UserKrb5CCacheName]} {
        if {[catch {file delete $UserKrb5CCacheName}]} {
        
            puts "Failed to destroy user credential cache $UserKrb5CCacheName."
            return false

        } 
    }
    return true
}



# Get all auth types are supported by web proxy server
proc GetProxyServerAuthType {} {
    global WebProxy
    global AuthTypeProxySupported
    global cip
    
    if {$WebProxy(Server) ne ""} {
        set cmd [CURL_CMD]
        append cmd " -x $WebProxy(Server) $cip -v"
        lassign [ExecuteCommand "$cmd 2>@1"] rc output
        
        foreach line [split $output "\n"] {
            if {[regexp -- {Proxy-Authenticate:\s+(\w+)\s*} $line -> authType]} {
                set authType [string tolower $authType]
                if {$authType eq [NAME_AUTH_NEGOTIATE]} {
                    set AuthTypeProxySupported(NAME_AUTH_NEGOTIATE) 1
                } elseif {$authType eq [NAME_AUTH_DIGEST]} {
                    set AuthTypeProxySupported(NAME_AUTH_DIGEST) 1
                } elseif {$authType eq [NAME_AUTH_NTLM]} {
                    set AuthTypeProxySupported(NAME_AUTH_NTLM) 1
                } elseif {$authType eq [NAME_AUTH_BASIC]} {
                    set AuthTypeProxySupported(NAME_AUTH_BASIC) 1
                }
            }
        }
    }
}

proc IsUsernameNtlmFormat {username} {
    if {[regexp {\\} $username] || [regexp {/} $username]} {
        return true
    }
    
    return false
}

proc IsUsernameUpnFormat {username} {
    if {[regexp {@} $username]} {
        return true
    }
    
    return false
}



proc GetHttpProxyConfig {} {
    global WebProxy
    global AuthTypeProxySupported
    set message ""
    set rc 0
    
    # get auth required and auth type.
    lassign [ExecuteCommand "[ADWEBPROXYCONF_CMD] -L 2>@1"] rc output
    
    # when adclient is down, we decide whether use configured http proxy
    # by seeing if the "Proxy Server:" configured is empty or not.
    if {[regexp {Proxy Server: (.*?)\n} $output -> proxyServer]} {
        
        if {$proxyServer ne ""} {
            set WebProxy(Server) $proxyServer
        }
    }
    
    if {$rc == 0 && [regexp {\nProxy Auth Required: (\w+)\nProxy Auth Type: (\w+)} $output -> authRequired authType]} {
    
        if {$authRequired eq "true"} {
            set WebProxy(AuthRequred) true
            set WebProxy(AuthType) $authType
        } else {
            set WebProxy(AuthType) [NAME_AUTH_NONE]
        }
        
    } else {
    
        if {$rc == [ADWEBPROXYCONF_ERROR_NO_FILE_FOUND]} {
            set message "Failed to access credential file."
            return [list $rc $message]
        } else {
            set message "Failed to get HTTP Proxy configuration."
            return [list $rc $message]
        }
        
    }
    
    # get auth user and credential
    lassign [ExecuteCommand "[ADWEBPROXYCONF_CMD]  -C -u 2>@1"] rc output
    
    if {$rc == 0} {

        set WebProxy(AuthUser) $output
        if {[IsMachineAccount $WebProxy(AuthUser)]} {
            set WebProxy(AuthType) [NAME_AUTH_NEGOTIATE]
        }

    } else {

        set message "Failed to access user name."
        return [list $rc $message]
        
    }
    
    GetProxyServerAuthType
    
    if {$WebProxy(AuthRequred) &&
        $WebProxy(AuthType) eq [NAME_AUTH_NEGOTIATE] ||
        $WebProxy(AuthType) eq [NAME_AUTH_ANY]} {
        
        # If "anyauth" is specified, curl will select one of auth types proxy supports, and the select order is "negotiate>digest>ntlm>basic"
        # The order is fixed and authentication is only done once that curl will not try to do ntlm again if negotiate fails, 
        # providing negotiate and ntlm are both supported by proxy server.
        if {$WebProxy(AuthType) eq [NAME_AUTH_ANY]} {
            if {$AuthTypeProxySupported(NAME_AUTH_NEGOTIATE) == 0} {
                if {$AuthTypeProxySupported(NAME_AUTH_NTLM) == 1} {
                    if {[IsUsernameUpnFormat $WebProxy(AuthUser)]} {
                        set message "The user name format \"$WebProxy(AuthUser)\" for 'ntlm' auth type is incorrect.\nPlease ensure the user name for 'ntlm' auth type is in the format of <username>, <domain>\\<username> or <domain>/<username>"
                        set rc [EXIT_ERR_WRONG_USER_NAME_FORMAT]
                    }
                }
                
                return [list $rc $message]
            }
        }
        
        if {[IsUsernameNtlmFormat $WebProxy(AuthUser)]} {
            set message "The user name format \"$WebProxy(AuthUser)\" for 'negotiate' auth type is incorrect.\nPlease ensure the user name for 'negotiate' auth type is in the format of <username> or <username>@<domain>"
            set rc [EXIT_ERR_WRONG_USER_NAME_FORMAT]
            return [list $rc $message]
        }
        

        lassign [ExecuteCommand "[ADWEBPROXYCONF_CMD] -C -p 2>@1"] rc output
        if {$rc != 0} {
            set message "Failed to decrypt user password, rc=$rc"
            return [list $rc $message]
        }
        

        set WebProxy(AuthKrb5CCache) [GetProxyKrb5CCache $WebProxy(AuthUser)]
        if {$WebProxy(AuthKrb5CCache) eq ""} {
            if {$WebProxy(AuthType) eq [NAME_AUTH_NEGOTIATE]} {
                set message "Wrong username/password."
            } else {
                set message "Failed to create KRB5 credential cache."
            }
            return [list [EXIT_ERR_KRB5CCACHE] $message]
        }

    }
    return [list $rc $message]

}


proc GetCIPAndProxyInfo {} {
    global WebProxy
    global cip
    set proxyServer ""
    set message ""
    set rc 0
    
    if {$cip ne ""} {
        # User specified --cip option, skip getting from adinfo.
        # directly connect to cip.
        set $WebProxy(DirectConnect) true
        return [list $rc $message]
    }
    
    
    lassign [ExecuteCommand "[ADINFO_CMD] --cipinfo"] rc output
    
    if {$rc == [ADINFO_ERR_CIP_NO_INFO]} {
        
        set message "Unable to get CIP info from adclient or kset"
        return [list [EXIT_ERR_NO_CIP_INFO] $message]
        
    } elseif {$rc == [ADINFO_ERR_NOT_JOINED]} {
    
        set message "Not joined to any domain"
        return [list [EXIT_ERR_NOT_JOINED] $message]
    
    } elseif {$rc != [ADINFO_ERR_CIP_DIRECT_CONNECT] && 
              $rc != [ADINFO_ERR_CIP_HTTP_PROXY_CONF] &&
              $rc != [ADINFO_ERR_CIP_CC_PROXY_ADCLIENT] &&
              $rc != [ADINFO_ERR_CIP_CC_PROXY_KSET]} {
    
        set message "unexpect return code from adinfo: $rc"
        return [list $rc $message]
        
    }
    
    if {[regexp {(.*),(.*)\n} $output -> url proxy]} {
    
        set cip $url
        set proxyServer $proxy
        
    } else {
    
        set message "Wrong CIP info format from adinfo"
        return [list [EXIT_ERR_NO_CIP_INFO] $message]
    
    }
    
    if {$rc == [ADINFO_ERR_CIP_DIRECT_CONNECT]} {
    
        # direct connection to cip
        set WebProxy(DirectConnect) "true"
        
    } elseif {$rc == [ADINFO_ERR_CIP_HTTP_PROXY_CONF]} {
    
        # connect to cip via configured http proxy
        
        # Get config from adwebproxyconf 
        lassign [GetHttpProxyConfig] rc message
        if {$rc} {
            return [list $rc $message]
        }
        
    } elseif {$rc == [ADINFO_ERR_CIP_CC_PROXY_ADCLIENT]} {
    
        # connetct to cip via connector returned from adclient
        set WebProxy(AuthType) [NAME_AUTH_NONE]
        
    } elseif {$rc == [ADINFO_ERR_CIP_CC_PROXY_KSET]} {
    
        # adclient is down. we have to decide the way connect to cloud url by ourselves.
        # since we cannot decide that by the return code of "adinfo --cipinfo".
        if {$proxyServer eq ""} {
        
            # direct connection to cip
            set WebProxy(DirectConnect) "true"
            
        } else {
        
            lassign [GetHttpProxyConfig] rc message
            if {$WebProxy(Server) eq ""} {
                
                # use connector.
                set WebProxy(AuthType) [NAME_AUTH_NONE]
                
            } else {
               
                # use configured http proxy.
                if {$rc != 0} {
                    return [list $rc $message]
                }
            }
        }
        
    } 
    
    set WebProxy(Server) $proxyServer
    
    return [list 0 $message]
}


#Prepare KRB5 CCache location for curl
proc SetCurlKrb5CCache {} {
    global WebProxy

    if {!$WebProxy(DirectConnect) && $WebProxy(Server) ne ""} {

        set ::env(KRB5CCNAME) $WebProxy(AuthKrb5CCache)
    } else {
        set ::env(KRB5CCNAME) ""
    }
    return ""
}


# escape special characters in password.
# special characters: \ $ ; # [] "
proc EscapeSpecialCharacters {password} {
    regsub -all -- {\\} $password {\\\\} password
    regsub -all -- {\$} $password {\\$} password
    regsub -all -- {\;} $password {\\;} password
    regsub -all -- {\#} $password {\\#} password
    regsub -all -- {\[} $password {\\[} password
    regsub -all -- {\]} $password {\\]} password
    regsub -all -- {\"} $password {\\"} password
    return $password
}

#
# Remove proxy credential information from curl command for output
# Input:    debugInfo
# Output:   modified debug info
#
proc FilterCurlCredentialInfo {debugInfo} {
    regsub -all -- {-U \"(.*?):(.*)\" } $debugInfo "-U \\1:*** " debugInfo
    return $debugInfo
}

# Prepare HTTP proxy parameters for curl
proc SetCurlHttpProxy {} {
    global WebProxy
    global AuthTypeProxySupported
    
    set curlproxy  ""
    if {!$WebProxy(DirectConnect) && $WebProxy(Server) ne ""} {

        append curlproxy " -x "
        append curlproxy $WebProxy(Server)
        
        # auth type is none means we use connector to connect to CIP
        if {$WebProxy(AuthType) ne [NAME_AUTH_NONE]} {
            append curlproxy " --proxy-" $WebProxy(AuthType)

            if {$WebProxy(AuthType) eq [NAME_AUTH_NEGOTIATE] ||
                ($WebProxy(AuthType) eq [NAME_AUTH_ANY] && 
                 $AuthTypeProxySupported(NAME_AUTH_NEGOTIATE) == 1)} {
                append curlproxy " -U :"
            } else {
                set cmd "[ADWEBPROXYCONF_CMD] -C -p"
                set password "[lindex [ExecuteCommand "$cmd 2>@1"] 1]"
                set password [EscapeSpecialCharacters $password]
                append curlproxy " -U \"" [EscapeSpecialCharacters $WebProxy(AuthUser)] ":" $password "\""
            }
        }
    }
    return $curlproxy
}

proc DownLoadSshPublicKey {} {
    global TestConnectionTimeout
    global cip
    global DownloadRESTAPI
    global OutPutFile

    set cmd "[SetCurlKrb5CCache][CURL_CMD]"
    append cmd "[SetCurlHttpProxy]"
    append cmd " --connect-timeout $TestConnectionTimeout --output $OutPutFile $cip/$DownloadRESTAPI"
    puts "Calling \"[FilterCurlCredentialInfo $cmd]\" to download..."
    lassign [ExecuteCommand "$cmd 2>@1"] rc output
    
    return [list $rc $output]
}

############################ main function ############################

lassign [ExecuteCommand "[ADID_CMD] -u"] status uid

if {$uid != 0} {
    puts "This command must be run as root.\n"
    exit [expr [EXIT_ERR_ROOT_PRIV]]
}

if {![GetOptions $argv]} {
    usage
    exit [expr [EXIT_ERR_USAGES]]
}

if {![info exists OutPutFile]} {
    puts "Option --output-file is not specified.\n"
    usage
    exit [expr [EXIT_ERR_USAGES]]
}

lassign [IsOutPutFileValid] rc message
if {$rc != 0} {
    puts $message
    exit [expr [EXIT_ERR_INVALID_OUTPUT_FILE]]
}

lassign [GetCIPAndProxyInfo] rc message
if {$rc != 0} {
    puts $message
    exit $rc
}

lassign [DownLoadSshPublicKey] rc message

DestroyUserKrb5CCache

if {$rc == 0} {
    puts "Download SSH Public Key to $OutPutFile successfully!"
    exec chmod 0644 $OutPutFile
    exit [EXIT_SUCCESS]
} else {
    puts "DownLoad SSH Public Key failed."
    exit $rc
}



