#!/bin/sh
# Copyright (C) 2015-2020 Centrify Corporation. All rights reserved.
#
# Collect info for all modified files in CDA pkg for dainfo --support
#
# Parameter: 'list': do 'ls -al' on the files changed/missing.
#            ''    : show files changed/missing.

CMD_VERIFY=""
CMD_AWK="awk"
PKG_NAME="CentrifyDA"

identify_os()
{
    UNAME=`uname -a`
    case "$UNAME" in
        AIX*)      
            CMD_VERIFY="pkgverify_aix"
            PKG_NAME="CentrifyDA.core"
            ;;
        HP-UX*)     
            CMD_VERIFY="pkgverify_hpux"
            ;;
        SunOS*)  
            CMD_VERIFY="pkgverify_sol"
            CMD_AWK="nawk"
            ;;
        Linux*)    
            if test -f /etc/debian_version; then
                PKG_NAME="centrifyda"
                CMD_VERIFY="pkgverify_deb"
            else
                CMD_VERIFY="pkgverify_rpm"
            fi
            ;;
        *)
            ;;
    esac
    readonly CMD_VERIFY PKG_NAME CMD_AWK
    export   CMD_VERIFY PKG_NAME CMD_AWK
}

# pkgverify routines, that: 
# if the checking passed, return 0, there is no output; 
# if the checking failed, return 1, the output mimick the output of 'rpm -V CentrifyDA'.

# pkgverify_rpm(): 
#   cmd: 'rpm -V', 
#   check: avail, checksum, size, mode, perm, device, link, owner, mtime, cap.  
#   if the checking failed, return 1, the output is as follows, e.g.:
#       missing     /etc/centrifyda/centrifyda.conf
#   or
#       S.5....T.  c /etc/centrifyda/centrifyda.conf
pkgverify_rpm()
{
    msg=`rpm -V ${PKG_NAME} 2>&1`
    if [ $? -eq 0 ]; then
        return 0
    fi 
    echo "$msg"
    return 1
}

# pkgverify_aix(): 
#   cmd: 'lppchk -c', 
#   check: avail, checksum, size.  
#   if the checking failed, return 1, the output of 'lppchk -c' is as follows, e.g.:
#       lppchk: 0504-206  File /etc/centrifyda/centrifyda.conf could not be located.
#   or
#       lppchk: 0504-208  Size of /etc/centrifyda/centrifyda.conf is 216, expected value was 217.
#   here the output is as follows, e.g.:
#       missing  /etc/centrifyda/centrifyda.conf
#   or
#       changed  /etc/centrifyda/centrifyda.conf
#
pkgverify_aix()
{
    msg=`lppchk -c ${PKG_NAME} 2>&1`
    if [ $? -eq 0 ]; then
        return 0
    fi  

    # it will never report the changes(unless removed) 
    # for files marked VOLATILE!
    # see install/lpp/fix_lpp_name.sh
    echo "$msg" \
        | grep -E '^lppchk: ' \
        | { \
            while read line; do
                file=`echo $line|sed -n 's/.* \(\/[^\ ]*\).*/\1/p'`
                # XXX "...located" must be in same line of 'File /path/file' ?
                # So far for all cdc files, it will be so, but not sure...
                if echo "$line" | grep -E ' could not be located\.$' >/dev/null 2>&1; then
                    echo "missing  $file"
                else
                    echo "changed  $file"
                fi
            done;
    }
    return 1
}

# pkgverify_deb(): 
#   cmd: 'dpkg -L/-s' 'md5sum', 
#   check: avail, checksum.
#   if the checking failed, here the output is as follows, e.g.:
#       missing  /etc/centrifyda/centrifyda.conf
#   or
#       changed  /etc/centrifyda/centrifyda.conf
#
pkgverify_deb()
{
    # dpkg stores checksums into two files: 
    #   otherfile:$pkg.md5sums and conffile:/var/lib/dpkg/status
    # the first one format is:
    #   hashval  /path/file 
    # the latter one format is:
    #   /path/file hashval

    mdfile="/var/lib/dpkg/info/${PKG_NAME}.md5sums"
    retno=0
    if [ -s "$mdfile" ]; then
        mkdir -p /var/centrifyda/
        hashfile=/var/centrifyda/${PKG_NAME}.md5sums
        cp $mdfile $hashfile
        cnfhash=`dpkg -s ${PKG_NAME} 2>&1`
        if [ $? -eq 0 ]; then
            echo "$cnfhash" \
                | ${CMD_AWK} '/Description:/{flag=0}flag;/Conffiles:/{flag=1}' \
                | ${CMD_AWK} '{print $2 " " $1}' \
                | sed 's/\///' \
                >> $hashfile
        fi
        files=`dpkg -L ${PKG_NAME}`
        for file in $files; do
            if [ ! -e "$file" ]; then
                echo "missing  $file"
                retno=1
                continue
            fi
            if [ -f "$file" ]; then
                rfile=`echo "$file" | cut -d/ -f 2-`
                phash=`grep -E "$rfile\$" "$hashfile" | cut -d\  -f 1`
                hash=`md5sum "$file" | cut -d\  -f 1`
                if [ -n "$phash" ] && [ "$hash" != "$phash" ]; then
                    echo "changed  $file"
                    retno=1
                fi
            fi
        done
        rm $hashfile
    fi
    return $retno
}

# pkgverify_hpux(): 
#   cmd: 'swverify -v' then 'swjob', 
#   check: avail, checksum, size, type, mtime, link, major/minor attributes.
#   if the checking failed, 'swverify' return 1; the output of 'swjob' is as follows, e.g.:
#       NOTE:    Volatile file "/etc/centrifyda/centrifyda.conf" missing.
#   or
#       ERROR:   File "/etc/centrifyda/centrifyda.conf" should have mtime
#                "1413638590" but the actual mtime is "1428376394".
#   here the output is as follows, e.g.:
#       missing  /etc/centrifyda/centrifyda.conf
#   or
#       changed  /etc/centrifyda/centrifyda.conf
#
pkgverify_hpux()
{
    msg=`swverify -v ${PKG_NAME} 2>&1`
    if [ $? -eq 0 ]; then
        return 0
    fi  
    # it will never report the changes(unless removed) 
    # for files marked VOLATILE!
    # see install/depot/replace-filelist-psf.sh: 'file -v'
    file=""
    newfile=""
    # we have to get the actual info from swjob.
    cmd=`echo "$msg"|grep command|sed -n 's/.* \"\(.*\)\".*/\1/p'`
    eval $cmd \
        | grep -E '^ERROR: *File |^NOTE: *Volatile file ' \
        | { \
            while read line; do
                # swverify may report multi issues for one file
                newfile=`echo "$line" | sed -n 's/.* \"\(\/[^\ ]*\)\".*/\1/p'`
                if echo "$line" | grep -E ' missing\.$' >/dev/null 2>&1; then
                    file=$newfile
                    echo "missing  $file"
                else
                    if [ ! "$newfile" = "$file" ]; then
                        file=$newfile
                        echo "changed  $file"
                    fi
                fi
            done;
    }

    return 1
}

# pkgverify_sol(): 
#   cmd: 'pkgchk' 
#   check: avail, checksum, size.  
#   if the checking failed, return 1, the output of 'pkgchk' is as follows, e.g.:
#       ERROR: /etc/centrifyda/centrifyda.conf
#           pathname does not exist
#   or
#       ERROR: /etc/centrifyda/centrifyda.conf
#           group name <sys> expected <root> actual
#           modtime <05/17/14 05:56:16 AM> expected <03/20/15 10:31:28 AM> actual
#           file size <112> expected <180> actual
#           file cksum <8570> expected <11809> actual
#   here the output is as follows, e.g.:
#       missing  /etc/centrifyda/centrifyda.conf
#   or
#       changed  /etc/centrifyda/centrifyda.conf
#
pkgverify_sol()
{
    msg=`pkgchk ${PKG_NAME} 2>&1`
    if [ $? -eq 0 ]; then
        return 0
    fi  
    # pkgchk may report multi issues for one file
    file=""
    newfile=""
    echo "$msg" \
        | { \
            while read line; do
                newfile=`echo "$line"|sed -n 's/^ERROR: \(.*\)/\1/p'`
                if [ -n "$newfile" ]; then
                    file="$newfile"
                elif echo "$line" | grep 'pathname does not exist' >/dev/null 2>&1; then
                    if [ -n "$file" ]; then 
                        echo "missing  $file"
                    fi
                    file=""
                else
                    if [ -n "$file" ]; then 
                        echo "changed  $file"
                    fi
                    file=""
                fi
            done
    }
    return 1
}

# main
identify_os
msg=`${CMD_VERIFY}`
ret=$?
if [ "$ret" -ne 0 -a -n "$msg" ]; then
    if [ "$1" = "list" ]; then
        if [ "$CMD_VERIFY" = "pkgverify_rpm" ]; then
            filelist=`echo "$msg" | ${CMD_AWK} '{if ($2 == "c") print $3; else print $2}'`
        else
            filelist=`echo "$msg" | ${CMD_AWK} '{print $2}'`
        fi
        if [ -n "$filelist" ]; then 
            ls -al $filelist 2>&1
        fi
    else
        echo "$msg"
    fi
fi
exit $ret
