#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2008-2020 Centrify Corporation. All rights reserved.
#
# Machine mapper script that configures user/group ignore settings.
#
#  The script get user/group ignore settings from group policy and write
#  them into centrifydc.conf (default) or ignore files.
#
#  If the GP template is CDC version 5.5 or earlier, or the populating
#  location is "centrifydcconf" (default), this GP script will stay the old
#  way, that is to modify centrifydc.conf.
#
#  If the GP template is CDC version 5.6 or later, and the populating
#  location is "ignorefile", this GP script will modify ignore files (the
#  path is default path which cannot be modified with this GP), and make sure
#  parameters nss.<user|group>.ignore in centrifydc.conf will be configured
#  as "file:/etc/centrifydc/<user|group>.ignore".
#
#  Note that, if users specified list of user/group ignore names, this GP
#  script will populate these user/group names into ignore files; if user
#  specified a file location, this GP script will restore local ignore
#  files.
#
#  Besides, pam.ignore.users is also updated when updating nss.user.ignore,
#  and make sure their values are same.
#
#  The syntax is one user name per line in ignore files:
#  user_name
#
#
#  Map:     modify centrifydc.conf or ignore files based on group policy
#           setting.
#              Not Configured: restore local setting (if it exists)
#              Enabled:        write gp setting into centrifydc.conf or
#                              ignore files
#              Disabled:       use default setting
#                              for ignore file, we just restore the local
#                              setting (if it exists) because it seems more
#                              reasonable than restoring the default file.
#                              Also, CDC version 5.5 or earlier will not
#                              touch the ignore file, we should keep the
#                              behavior consistent.
#
#  Unmap:   restore local setting (if it exists)
#
#
# Parameters: <map|unmap> mode
#   map|unmap   action to take
#   mode        mode (not used)
#
#
# Exit value:
#   0   Normal
#   1   Error
#   2   usage
#
##############################################################################

use strict;
use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;
use CentrifyDC::GP::Args;
use CentrifyDC::GP::RegHelper;
use CentrifyDC::GP::Registry;
use CentrifyDC::GP::Mapper;
use CentrifyDC::GP::General qw(:debug IsEmpty ReadFile WriteFile CreateFile CopyFile);

# comment line at the beginning of ignore files
my @MARKER = (
    '######################################################################',
    '#        DO NOT EDIT. CentrifyDC group policy generated file.        #',
    '#       Changes to this file can be made through group policy.       #',
    '######################################################################',
);

my @IGNORE_TYPES = qw(user group);

my %NSS_IGNORE = (
    user => "nss.user.ignore",
    group => "nss.group.ignore",
);

my $PAM_IGNORE_USER = "pam.ignore.users";

my %NSS_IGNORE_DEFAULT = (
    user => "nss_user_ignore_default",
    group => "nss_group_ignore_default",
);

my %LOCATION_SWITCH = (
    user => "user_ignore_location_switch",
    group => "group_ignore_location_switch",
);

my %IGNORE_FILE_PATH = (
    user => "/etc/centrifydc/user.ignore",
    group => "/etc/centrifydc/group.ignore",
);

my %IGNORE_LOCAL_FILE_PATH = (
    user => "/var/centrifydc/reg/user.ignore.save",
    group => "/var/centrifydc/reg/group.ignore.save",
);

my $file;

my $REGKEY = "Software/Policies/Centrify/CentrifyDC/Settings/Login";

my $CENTRIFYDC_CONF_FLAG     = "centrifydcconf";
my $IGNORE_FILE_FLAG         = "ignorefile";

# hash of registry setting
my $registry_setting;

sub ::DoAdFlush();
sub UpdateGPForCentrifyDCConf($);
sub UpdateGPForIgnoreFile($);
sub SaveLocalIgnoreFile($);
sub RestoreLocalIgnoreFile($);
sub DoIgnoreFileMap($);
sub DoIgnoreFileUnmap();

# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#
# create file for adreload/adflush
#
sub ::DoAdFlush()
{
    CreateFile("/var/centrifydc/reg/do_adflush");
    CreateFile("/var/centrifydc/reg/do_adreload"); 
}

#
# Update generic map setting for populating to centrifydc.conf
#
# Note that the map setting of pam.ignore.users is the same as nss.user.ignore.
#
#   $_[0]:  type    (user/group)
#
sub UpdateGPForCentrifyDCConf($)
{
    my $type = $_[0];

    $file->{value_map}->{$NSS_IGNORE{$type}} = {
      'active' => '$value_map->{' . $LOCATION_SWITCH{$type} . '}{reg_data} ne "ignorefile"',
      'default_data' => 'file:/etc/centrifydc/user.ignore',
      'reg_class'  => 'machine',
      'reg_key'    => $REGKEY,
      'reg_type'   => [ 'REG_SZ', ],
      'reg_value'  => $NSS_IGNORE{$type},
      'value_type' => 'named',
    };
   
    if ($type eq "user")
    {
        $file->{value_map}->{$PAM_IGNORE_USER} = {
          'active' => '$value_map->{' . $LOCATION_SWITCH{$type} . '}{reg_data} ne "ignorefile"',
          'default_data' => 'file:/etc/centrifydc/user.ignore',
          'reg_class'  => 'machine',
          'reg_key'    => $REGKEY,
          'reg_type'   => [ 'REG_SZ', ],
          'reg_value'  => $NSS_IGNORE{$type},
          'value_type' => 'named',
        };
    }
}

#
# Update generic map setting for populating to ignore file, make sure
# parameters nss.<user|group>.ignore is in centrifydc.conf
# will be configured as "file:/etc/centrifydc/<user|group>.ignore".
#
# Note that the map setting of pam.ignore.users is the same as nss.user.ignore.
#
#   $_[0]:  type    (user/group)
#
sub UpdateGPForIgnoreFile($)
{
    my $type = $_[0];

    $file->{value_map}->{$NSS_IGNORE{$type}} = {
      'active' => '$value_map->{' . $LOCATION_SWITCH{$type} . '}{reg_data} eq "ignorefile"',
      'data_value' => $NSS_IGNORE_DEFAULT{$type},
      'default_data' => 'file:/etc/centrifydc/user.ignore',
      'reg_class' => 'machine',
      'reg_key' => $REGKEY,
      'reg_type' => [ 'REG_SZ', ],
      'reg_value' => $NSS_IGNORE{$type},
      'value_type' => 'named',
    };
   
    if ($type eq "user")
    {
        $file->{value_map}->{$PAM_IGNORE_USER} = {
          'active' => '$value_map->{' . $LOCATION_SWITCH{$type} . '}{reg_data} eq "ignorefile"',
          'data_value' => $NSS_IGNORE_DEFAULT{$type},
          'default_data' => 'file:/etc/centrifydc/user.ignore',
          'reg_class' => 'machine',
          'reg_key' => $REGKEY,
          'reg_type' => [ 'REG_SZ', ],
          'reg_value' => $NSS_IGNORE{$type},
          'value_type' => 'named',
        };
    }
}

#
# Save the local ignore file.
#
#   $_[0]:  type    (user/group)
#
sub SaveLocalIgnoreFile($)
{
    my $type = $_[0];
    my $ret;

    if (! -e $IGNORE_LOCAL_FILE_PATH{$type})
    {
        DEBUG_OUT("Save local ignore file to $IGNORE_LOCAL_FILE_PATH{$type}");
        if (-e $IGNORE_FILE_PATH{$type})
        {
            $ret = CopyFile($IGNORE_FILE_PATH{$type}, $IGNORE_LOCAL_FILE_PATH{$type});
            if (!defined($ret) or $ret != 1)
            {
                FATAL_OUT("Cannot save local ignore file to $IGNORE_LOCAL_FILE_PATH{$type}");
            }
        }
        else
        {
            # almost never happens.
            DEBUG_OUT("$IGNORE_FILE_PATH{$type} doesn't exist.");
        }
    }
}

#
# Restore the local ignore file.
#
#   $_[0]:  type    (user/group)
#
sub RestoreLocalIgnoreFile($)
{
    my $type = $_[0];
    my $ret;

    if (-e $IGNORE_LOCAL_FILE_PATH{$type})
    {
        DEBUG_OUT("Restore local ignore file from $IGNORE_LOCAL_FILE_PATH{$type}");
        $ret = CopyFile($IGNORE_LOCAL_FILE_PATH{$type}, $IGNORE_FILE_PATH{$type});
        if (!defined($ret) or $ret != 1)
        {
            FATAL_OUT("Cannot restore local ignore file from $IGNORE_LOCAL_FILE_PATH{$type}");
        }
        else
        {
            unlink($IGNORE_LOCAL_FILE_PATH{$type}) || FATAL_OUT $!;
        }
    }            
}

#
# Map user/group to ignore file according to the GP setting.
#
#  - if GP template is CDC version 5.6 or later and the populating location
#    is "ignorefile".
#    - if users specified a file location, restore local ignore files.
#    - if users specified list of user/group ignore names, populate these
#      user/group names into default ignore files.
#  - if GP template is CDC version 5.5 or earlier, or the populating location
#    is "centrifydcconf".
#    - restore local ignore files.
#
#   $_[0]:  action  (map/unmap)
#
sub DoIgnoreFileMap($)
{
    my $action = $_[0];
    my $changed = 0;

    foreach my $type (@IGNORE_TYPES)
    {
        my $to_file = ($registry_setting->{$type}->{location_cur} eq $IGNORE_FILE_FLAG);
        DEBUG_OUT("Populate ignore %ss to %s.", $type, $to_file ? $IGNORE_FILE_PATH{$type} : "centrifydc.conf");

        my $reg_ignores = CentrifyDC::GP::RegHelper->new($action, 'machine', $REGKEY, $NSS_IGNORE{$type}, undef);
        $reg_ignores or return undef;
        $reg_ignores->load();
        my $setting = $reg_ignores->get("current");

        if ($to_file)
        {
            if ($setting =~ /^file:*/)
            {
                # The setting is a file reference, nothing we can populate
                # to ignore file. Restore the ignore file in case this file
                # reference is that ignore file.
                RestoreLocalIgnoreFile($type);
            }
            else
            {
                # Write setting to ignore file, save the local ignore file
                # before overwriting it.
                SaveLocalIgnoreFile($type);
         
                DEBUG_OUT("User specifies some %s names, populate to %s ignore file.", $type, $type);
                
                $setting =~ s/^\s+//;
                $setting =~ s/\s+$//;
                $setting =~ s/\s+/\n/g;
                my $str = join("\n", @MARKER);
                $str .= "\n" . $setting . "\n";
                my $sys_string = ReadFile($IGNORE_FILE_PATH{$type});

                # If the ignore file content is undef, or it is not the same as the setting
                if (!defined($sys_string) || $sys_string ne $str)
                {
                    DEBUG_OUT("Write %s ignore file " . $IGNORE_FILE_PATH{$type}, $type);
                    unlink($IGNORE_FILE_PATH{$type}) || FATAL_OUT $!;
                    my $rc = WriteFile($IGNORE_FILE_PATH{$type}, $str, {'mode' => 0644});
                    if (! $rc)
                    {
                        ERROR_OUT("Cannot write %s ignore file " . $IGNORE_FILE_PATH{$type}, $type);
                        return;
                    }
                    $changed = 1;
                }
            }
        }
        else
        {
            # Populating to centrifydc.conf, unconfigured or disabled,
            # we just restore the local ignore file in these cases.
            RestoreLocalIgnoreFile($type);
        }
    }

    if ($changed)
    {
        DEBUG_OUT("User/Group ignore file content changed. Need to do adflush.");
        DoAdFlush();
    }
}

#
# Unmap user/group from ignore file.
#
sub DoIgnoreFileUnmap()
{
    foreach my $type (@IGNORE_TYPES)
    {
        # Restore local ignore file
        RestoreLocalIgnoreFile($type);
    }
}

# >>> DATA >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

$file = {
    'comment_markers' => [
      '#',
    ],
    'hierarchy_separator' => '.',
    'list_expr' => ', *| +',
    'list_separator' => ', ',
    'match_expr' => [
      '/^\s*([^\s:=]+)[:=]\s*(.*)/',
    ],
    'parent_expr' => '^(.*)\.([^\.]+)$',
    'path' => [
      '/etc/centrifydc/centrifydc.conf',
    ],
    'lock' => '/etc/centrifydc/centrifydc.conf.lock',
    'post_command' => '::DoAdFlush()',
    'value_map' => {
      'user_ignore_location_switch' => {
        'active' => '0',
        'reg_class' => 'machine',
        'reg_key' => "$REGKEY",
        'reg_type' => [
          'REG_SZ',
        ],
        'reg_value' => 'user_ignore_location_switch',
        'value_type' => 'named',
      },
      'group_ignore_location_switch' => {
        'active' => '0',
        'reg_class' => 'machine',
        'reg_key' => "$REGKEY",
        'reg_type' => [
          'REG_SZ',
        ],
        'reg_value' => 'group_ignore_location_switch',
        'value_type' => 'named',
      },
    },
    'write_data' => '$value: $data\n',
};

$registry_setting = {
    'user' => {
        'location_cur' => "$CENTRIFYDC_CONF_FLAG",
        'location_pre' => "$CENTRIFYDC_CONF_FLAG",
    },
    'group' => {
        'location_cur' => "$CENTRIFYDC_CONF_FLAG",
        'location_pre' => "$CENTRIFYDC_CONF_FLAG",
    },
};

# >>> MAIN >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $args = CentrifyDC::GP::Args->new('machine');

CentrifyDC::GP::Registry::Load(undef);

# Check the pupulating locations of the previous and current GP settings.
# Note that there is no need to check whether GP is configured, enabled
# or disabled because:
# - For updating ignore file, we just care about the location
# - For updating centrifydc.conf, this is actualy done by generic GP
#
foreach my $type (@IGNORE_TYPES)
{
    my $cur = (CentrifyDC::GP::Registry::Query($args->class(), $REGKEY, "current", $LOCATION_SWITCH{$type}))[1];
    my $pre = (CentrifyDC::GP::Registry::Query($args->class(), $REGKEY, "previous", $LOCATION_SWITCH{$type}))[1];
    if (defined $cur)
    {
        $registry_setting->{$type}->{location_cur} = $cur;
    }
    if (defined $pre)
    {
        $registry_setting->{$type}->{location_pre} = $pre;
    }
}

if ($args->isMap())
{
    # For ignore file
    DoIgnoreFileMap($args->action());

    # For centrifydc.conf

    # Since the map setting ($file) for the same parameters
    # (nss.<user|group>.ignore and pam.ignore.users) are different under
    # different populating locations, so after user switched the populating
    # location, we must perform the previous map setting so that it gets
    # a chance to update with the current location setting. After that, we
    # can change the map setting and map with the current location setting.
    #
    my $do_pre_map = 0;
    foreach my $type (@IGNORE_TYPES)
    {
        if ($registry_setting->{$type}->{location_pre} ne $registry_setting->{$type}->{location_cur})
        {
            if ($registry_setting->{$type}->{location_pre} eq $IGNORE_FILE_FLAG)
            {
                UpdateGPForIgnoreFile($type);
            }
            else
            {
                UpdateGPForCentrifyDCConf($type);
            }
            $do_pre_map = 1;
        }
    }
    if ($do_pre_map)
    {
        DEBUG_OUT("Populating location changed, update with previous map setting first.");
        CentrifyDC::GP::Mapper::Map($file, $args->user());
    }

    foreach my $type (@IGNORE_TYPES)
    {
        if ($registry_setting->{$type}->{location_cur} eq $IGNORE_FILE_FLAG)
        {
            UpdateGPForIgnoreFile($type);
        }
        else
        {
            UpdateGPForCentrifyDCConf($type);
        }
    }
    DEBUG_OUT("Do generic map.");
    CentrifyDC::GP::Mapper::Map($file, $args->user());
}
else
{
    # For ignore file
    DoIgnoreFileUnmap();

    # For centrifydc.conf
    foreach my $type (@IGNORE_TYPES)
    {
        if ($registry_setting->{$type}->{location_cur} eq $IGNORE_FILE_FLAG)
        {
            UpdateGPForIgnoreFile($type);
        }
        else
        {
            UpdateGPForCentrifyDCConf($type);
        }
    }
    DEBUG_OUT("Do generic unmap.");
    CentrifyDC::GP::Mapper::UnMap($file, $args->user());
}

